/*	    _   __ ___    _____    ______ __  __ ____
   	   / | / //   |  / ___/   / ____// / / //  _/
  	  /  |/ // /| |  \__ \   / / __ / / / / / /
 	 / /|  // ___ | ___/ /  / /_/ // /_/ /_/ /
	/_/ |_//_/  |_|/____/   \____/ \____//___/
			Not Another Skinnable GUI

	NAS AMP Media Player by Steve Terry
	June 28, 2003
*/

#include <allegro.h>
#include <nas.h>
#include <string.h>
#include <ctype.h>
#include "nas_amp.h"

NAS_WIDGET *nas_amp = NULL;

typedef struct amp_item{
	char *name;
	char *title;
	int seconds;
	int status;
}AMP_ITEM;

typedef struct spectrum_state{
	int num_bars;
	int falloff_rate;
}SPECTRUM_STATE;

static float *spectrum;
static float s[512];
static FSOUND_STREAM *stream = NULL;
static FMUSIC_MODULE *mod = NULL;
static FSOUND_DSPUNIT *OscUnit = NULL;
static signed short *OscBuffer = NULL;
static int OscBlock = 0;
static int fsound_cd = 0;
static char fsound_cd_drive = 0;
static int fsound_cd_track = 1;
static int  metanum = 0;
static char net_artist[256] = "";	
static char net_title[256]  = "";
static int fsound_net_stream = 0;

static int kbps = 0;

static unsigned char paused = 0;
static signed char playing = FALSE;
static unsigned char repeat = FALSE;
static unsigned char shuffle = FALSE;

//static int tracker = 0;
static int curstatus = -1;
static int numplayed = 0;

/*static MENU amp_about_menu[] =
{
   { "N.A.S. AMP v.91",   NULL, NULL, 0, NULL },
   { NULL,                NULL, NULL, 0, NULL }
};

static MENU amp_menu[] =
{
   { "&File", 	NULL, NULL,   	      0, NULL },
   { "&About",  NULL, amp_about_menu, 0, NULL },
   { NULL,      NULL, NULL,           0, NULL }
};*/

static AMP_ITEM *item_playing = NULL;
static int old_kbps = 0;

static NAS_WIDGET *amp_prev;
static NAS_WIDGET *amp_next;
static NAS_WIDGET *amp_play;
static NAS_WIDGET *amp_pause;
static NAS_WIDGET *amp_stop;
static NAS_WIDGET *amp_eject;
static NAS_WIDGET *amp_shuffle;
static NAS_WIDGET *amp_repeat;
static NAS_WIDGET *amp_progress;
static NAS_WIDGET *amp_volume;
static NAS_WIDGET *amp_group;
static NAS_WIDGET *amp_spect;
static NAS_WIDGET *amp_time_text;
static NAS_WIDGET *amp_time_box;
static NAS_WIDGET *amp_cur_time;
static NAS_WIDGET *amp_kbps_text;
static NAS_WIDGET *amp_kbps_box;
static NAS_WIDGET *amp_cur_kbps;
static NAS_WIDGET *amp_list;
static NAS_WIDGET *amp_scroll_box;
static NAS_WIDGET *amp_scroll_text;

static int nas_spectrum_proc(int msg, NAS_WIDGET *d, BITMAP *wnd, int c);
static int nas_load_proc(int msg, NAS_WIDGET *d, BITMAP *wnd, int c);

static NAS_WIDGET *nas_add_spectrum(NAS_WIDGET *d, int x, int y, int w, int h, U_LINT flags, int num_bars, int falloff_rate);

static int ProgressSetPosCallback(NAS_WIDGET *d, int max, int pos, void *dp);
static int PlayerProgressCallback(NAS_WIDGET *d, void *dp);
static int LoadfileCallback(NAS_WIDGET *d, int msg, LIST_ITEM *item, void *dp);
static int PlayCallback(NAS_WIDGET *d, void *dp);
static int StopCallback(NAS_WIDGET *d, void *dp);
static int CloseCallback(NAS_WIDGET *d, void *dp);
static int PauseCallback(NAS_WIDGET *d, void *dp);
static int RepeatCallback(NAS_WIDGET *d, void *dp);
static int ShuffleCallback(NAS_WIDGET *d, void *dp);
static int NextCallback(NAS_WIDGET *d, void *dp);
static int PrevCallback(NAS_WIDGET *d, void *dp);
static int EjectCallback(NAS_WIDGET *d, void *dp);
static int VolumeCallback(NAS_WIDGET *d, int max, int pos, void *dp);
static int DialogCallback(NAS_WIDGET *d, int msg, int *w, int *h);
static int GroupboxCallback(NAS_WIDGET *d, void *dp);
static void DestroyCallback(LIST_ITEM *item);
static void DrawCallback(BITMAP *dest, int x, int y, LIST_ITEM *item);

static signed char F_CALLBACKAPI metacallback(char *name, char *value, int userdata);
static void *F_CALLBACKAPI OscCallback(void *originalbuffer, void *newbuffer, int length, int param);

static void PlotOscilliscope(BITMAP *dest, int flags, int x, int y, int w, int h);
static void PlotSpectrum(BITMAP *dest, int num_bars, int falloff_rate, int flags, int x, int y, int w, int h);
static void draw_bar_gradient(BITMAP *dest, int x, int y, int w, int h, int h2, int col1, int col2);

static int load_m3u(const char *path);
static int load_pls(const char *path);

static void make_title(AMP_ITEM **amp);

static int get_kbps(FSOUND_STREAM *stream);

static void DestroyCallback(LIST_ITEM *item){
	AMP_ITEM *amp = (AMP_ITEM *)item->userobj;
	free(amp->name);
	if(amp->title)
		free(amp->title);
}

static signed char F_CALLBACKAPI metacallback(char *name, char *value, int userdata){
    	if (!ustrcmp("ARTIST", name)){
       		ustrcpy(net_artist, value);
        	return TRUE;
   	}
    	if (!ustrcmp("TITLE", name)){
        	ustrcpy(net_title, value);
        	metanum++;
        	return TRUE;
    	}
    	return TRUE;
}

static void make_title(AMP_ITEM **amp){
	FILE *fin;
	ID3Tag info;
	char buffer[256];
	char *tmp;
	char *fname;
	fname = get_filename((*amp)->name);
	if((fin = fopen((*amp)->name, "rb+")) == NULL){
		if ((!ustrncmp((*amp)->name, "http://", 7) || !ustrncmp((*amp)->name, "http:\\\\", 7))){
			return;
		}
		else{
			nasmouse_set_mouse_sprite(NASMOUSE_WAIT);
			// failed to open.. possibly a CD or invalid file
			tmp = gettitle(fname);
			if(tmp[0] == '\0')
				usprintf(buffer, "%s", getartist(fname));
			else
				usprintf(buffer, "%s - %s", getartist(fname), gettitle(fname));
			(*amp)->title = (char*)malloc(sizeof(char) * ustrlen(buffer) + 1);
			ustrcpy((*amp)->title, buffer);
			nasmouse_set_mouse_sprite(NASMOUSE_ARROW);
		}
		return;
	}
	else
		nasmouse_set_mouse_sprite(NASMOUSE_WAIT);
	getTag(fin, &info);
	if(info.Artist[0] == '\0'){
		tmp = gettitle(fname);
		if(tmp[0] == '\0')
			usprintf(buffer, "%s", getartist(fname));
		else
			usprintf(buffer, "%s - %s", getartist(fname), gettitle(fname));
		(*amp)->title = (char*)malloc(sizeof(char) * ustrlen(buffer) + 1);
		ustrcpy((*amp)->title, buffer);
	}
	else{
		usprintf(buffer, "%s - %s", info.Artist, info.Title);
		(*amp)->title = (char*)malloc(sizeof(char) * ustrlen(buffer) + 1);
		ustrcpy((*amp)->title, buffer);
	}
	fclose(fin);
	nasmouse_set_mouse_sprite(NASMOUSE_ARROW);
}

static void DrawCallback(BITMAP *dest, int x, int y, LIST_ITEM *item){
	AMP_ITEM *amp = (AMP_ITEM *)item->userobj;
	if(!amp->title)
		make_title(&amp);
	nas_listbox_textout(dest, x, y, amp->title ? amp->title : amp->name, item);
}

static int load_m3u(const char *path){
	AMP_ITEM item;
	char *header = "#EXTM3U";
	char *id = "#EXTINF";
	char *tmp, *tmp2;
	FILE *fin;
	char buffer[256];
	if((fin = fopen(path, "r")) == NULL)
		return 0;
	fscanf(fin, " %[^\n] ", buffer);
	if(ustrcmp(buffer, header) != 0){
		fclose(fin);
		return 0;
	}
	while(fscanf(fin, " %[^\n] ", buffer) != EOF){
		tmp = ustrtok(buffer, ":");
		if(ustrcmp(tmp, id) != 0){
			tmp2 = ustrtok(NULL, "");
			if(tmp2){
				usprintf(buffer, "%s:%s", tmp, tmp2);	// Fix the path back to it's origional state
				tmp = buffer;
			}
			item.name = (char *)malloc(sizeof(char) * ustrlen(tmp) + 1);
			ustrcpy(item.name, tmp);
			item.title = NULL;
			item.status = -1;
			item.seconds = 0;
			nas_listbox_insert(amp_list, &item, sizeof(item), 100, nasfont_text_height(AL_font, nas_shadow_mode(listitem.textcol, 4)));
		}
		else{
			tmp = ustrtok(NULL, ",");
			item.seconds = atoi(tmp);
			tmp = ustrtok(NULL, "");
			item.title = (char *)malloc(sizeof(char) * ustrlen(tmp) + 1);
			ustrcpy(item.title, tmp);
			fscanf(fin, " %[^\n] ", buffer);
			item.name = (char *)malloc(sizeof(char) * ustrlen(buffer) + 1);
			ustrcpy(item.name, buffer);
			item.status = -1;
			item.seconds = 0;
			nas_listbox_insert(amp_list, &item, sizeof(item), 100, nasfont_text_height(AL_font, nas_shadow_mode(listitem.textcol, 4)));
		}
	}
	fclose(fin);
	return 1;
}

static int load_pls(const char *path){
	AMP_ITEM item;
	const char *buffer;
	char section[256];
	int i, entries;
	push_config_state();
	set_config_file(path);
	if(get_config_int("playlist", "Version", 0) != 2){
		pop_config_state();
		return 0;
	}
	if((entries = get_config_int("playlist", "NumberOfEntries", 0)) == 0){
		pop_config_state();
		return 0;
	}
	for(i = 0; i < entries; i++){
		usprintf(section, "File%d", i + 1);
		if((buffer = get_config_string("playlist", section, NULL)) == NULL){
			pop_config_state();
			return 0;
		}
		item.name = (char *)malloc(sizeof(char) * (ustrlen(buffer) + 1));
		ustrcpy(item.name, buffer);
		usprintf(section, "Title%d", i + 1);
		buffer = get_config_string("playlist", section, NULL);
		if(buffer){
			item.title = (char *)malloc(sizeof(char) * (ustrlen(buffer) + 1));
			ustrcpy(item.title, buffer);
		}
		else
			item.title = NULL;
		usprintf(section, "Length%d", i + 1);
		buffer = get_config_string("playlist", section, NULL);
		if(buffer)
			item.seconds = atoi(buffer);
		else
			item.seconds = 0;
		item.status = -1;
		nas_listbox_insert(amp_list, &item, sizeof(item), 100, nasfont_text_height(AL_font, nas_shadow_mode(listitem.textcol, 4)));
	}
	pop_config_state();
	return 1;
}

static int nas_load_proc(int msg, NAS_WIDGET *d, BITMAP *wnd, int c){
	AMP_ITEM *item;
	LIST_ITEM *tmp;
	int ret = D_O_K;
	int i, rnd, listitems;
	if(msg == MSG_START){
                const char *path;
		push_config_state();
		set_config_file("amp.cfg");
		path = get_config_string("AMP", "path", NULL);
		if(path)
			if(!load_m3u(path))
				load_pls(path);
		listitems = nas_get_listbox_numitems(amp_list);
		for(i = 0; i < listitems; i++){
			do{
				rnd = rand() % listitems;
				tmp = nas_get_listbox_item(amp_list, rnd);
				item = tmp->userobj;
			}while(item->status != -1);
			item->status = i;	
		}
		pop_config_state();
	}
	else if(msg == MSG_END){
		// Listbox doesn't clear the list upon MSG_END, explicitly do so.
		nas_listbox_clear(amp_list);
	}
	return ret;
}

int init_amp(){
	NAS_WIDGET *file_menu;
	NAS_WIDGET *about_menu;
	int buff, freq, ch;
	const char *tmp;
	time_t t;
	time(&t);
	srand(t);
	push_config_state();
	set_config_file("amp.cfg");
	buff = get_config_int("AMP", "buffer", AMP_DEFAULT_BUFFER);
	freq = get_config_int("AMP", "frequency", AMP_DEFAULT_FREQ);
	ch = get_config_int("AMP", "maxchannels", AMP_DEFAULT_CHAN);
	tmp = get_config_string("AMP", "drive", "0");
	if(tmp[0] == '0')
		fsound_cd_drive = 0;
	else
		fsound_cd_drive = tmp[0];
	FSOUND_SetBufferSize(buff);
   	if(!FSOUND_Init(freq, ch, FSOUND_INIT_GLOBALFOCUS))
		return 0;
	FSOUND_DSP_SetActive(FSOUND_DSP_GetFFTUnit(), TRUE);
	OscUnit	= FSOUND_DSP_Create(&OscCallback, FSOUND_DSP_DEFAULTPRIORITY_USER + 3, 0);
	OscBuffer = calloc(FSOUND_DSP_GetBufferLengthTotal() + 16, 2); /* *2 for mono 16bit buffer */
	pop_config_state();
	nas_amp = nas_create_dialog(100, 100, 600, 600, AMP_DIALOG_FLAGS, "NAS Amp");
	amp_prev = nas_add_button(nas_amp, 160, 264, 40, 40, 'b', AMP_BUTTON_FLAGS, "&Prev", PrevCallback, NULL);
	amp_next = nas_add_button(nas_amp, 220, 264, 40, 40, 'n', AMP_BUTTON_FLAGS, "&Next", NextCallback, NULL);
	amp_play = nas_add_button(nas_amp, 280, 264, 40, 40, 'p', AMP_BUTTON_FLAGS, "&Play", PlayCallback, NULL);
	amp_pause = nas_add_button(nas_amp, 340, 264, 40, 40, 'a', AMP_BUTTON_FLAGS, "P&ause", PauseCallback, NULL);
	amp_stop = nas_add_button(nas_amp, 400, 264, 40, 40, 's', AMP_BUTTON_FLAGS, "&Stop", StopCallback, NULL);
	amp_eject = nas_add_button(nas_amp, 460, 264, 40, 40, 'e', AMP_BUTTON_FLAGS, "&Eject", EjectCallback, NULL);
	amp_shuffle = nas_add_checkbox(nas_amp, 120, 220, 80, 20, 'h', AMP_DEFLT_FLAGS, "S&huffle", ShuffleCallback, NULL);
	amp_repeat = nas_add_checkbox(nas_amp, 120, 235, 80, 20, 'r', AMP_DEFLT_FLAGS, "&Repeat", RepeatCallback, NULL);
	amp_progress = nas_add_progress(nas_amp, 120, 210, 172, 18, 0, 0, 255, "0%");
	amp_volume = nas_add_slider(nas_amp, 290, 135, 8, 64, 0, 0, 255);
	amp_group = nas_add_groupbox(nas_amp, 2, 2, 168, 68, 0, "Spectrum");
	amp_spect = nas_add_spectrum(amp_group, 4, 2, 160, 64, AMP_SPECT_FLAGS, 50, 15);
	amp_time_text = nas_add_text(nas_amp, 120, 200, 40, 16, AMP_DEFLT_FLAGS, NAS_TEXT_ALIGN_LEFT, "Time:");
	amp_time_box = nas_add_box(nas_amp, 120, 200, 40, 16, NAS_BOX_SUNKEN);
	amp_cur_time = nas_add_text(nas_amp, 120, 200, 40, 16, D_DEFAULT_H, NAS_TEXT_ALIGN_LEFT, "00:00.00");
	amp_kbps_text = nas_add_text(nas_amp, 120, 200, 40, 16, AMP_DEFLT_FLAGS, NAS_TEXT_ALIGN_LEFT, "KBPS:");
	amp_kbps_box = nas_add_box(nas_amp, 120, 200, 40, 16, NAS_BOX_SUNKEN);
	amp_cur_kbps = nas_add_text(nas_amp, 120, 200, 40, 16, D_DEFAULT_H, NAS_TEXT_ALIGN_RIGHT, "0");
	amp_list = nas_add_listbox(nas_amp, 310, 158, 220, 84, 0, LoadfileCallback, NULL);
	amp_scroll_box = nas_add_box(nas_amp, 200, 200, 220, 20, NAS_BOX_SUNKEN);
	amp_scroll_text = nas_add_textscroll(nas_amp, 200, 200, 220, 20, 0, "N.A.S. Amp v.91");
	nas_add_widget(nas_amp, nas_load_proc, 0, 0, 0, 0, 0, 0, NULL);
	nas_set_dialog_callback(nas_amp, DialogCallback);
	nas_set_dialog_close_callback(nas_amp, CloseCallback, NULL);
	nas_set_dialog_icon_ex(nas_amp, "./amp.bmp");
	file_menu = nas_add_dialog_menu(nas_amp, 'f', "&File");
	about_menu = nas_add_dialog_menu(nas_amp, 'a', "&About");
	nas_add_menu_item(about_menu, 0, "N.A.S. AMP v.91");
	nas_add_menu_item(file_menu, 'o', "&Open Playlist");
	nas_add_menu_item(file_menu, 's', "&Save Playlist");
	nas_add_menu_item(file_menu, 0, NULL);
	nas_add_menu_item(file_menu, 'e', "&Exit");
	nas_set_groupbox_callback(amp_group, GroupboxCallback, NULL);
	nas_set_listbox_draw_callback(amp_list, DrawCallback);
	nas_set_listbox_destroy_callback(amp_list, DestroyCallback);
	nas_set_progress_click_callback(amp_progress, ProgressSetPosCallback, NULL);
	nas_set_progress_idle_callback(amp_progress, PlayerProgressCallback, NULL);
	nas_set_slider_callback(amp_volume, VolumeCallback, NULL);
	nas_set_textscroll_speed(amp_scroll_text, 1);
	return 1;
}

static NAS_WIDGET *nas_add_spectrum(NAS_WIDGET *d, int x, int y, int w, int h, U_LINT flags, int num_bars, int falloff_rate){
	SPECTRUM_STATE *tmp;
	NAS_WIDGET *tmp2 = nas_add_widget(d, nas_spectrum_proc, x, y, w, h, 0, flags, NULL);
	if(!tmp2)
		return NULL;
	tmp = (SPECTRUM_STATE*)calloc(1, sizeof(SPECTRUM_STATE));
	if(!tmp)
		return NULL;
	tmp->num_bars = num_bars;
	tmp->falloff_rate = falloff_rate;
	tmp2->dp = tmp;
	return tmp2;
}

void amp_exit(){
	if(stream)
		FSOUND_Stream_Close(stream);
	if(mod)
		FMUSIC_FreeSong(mod);
	if (OscUnit){
		FSOUND_DSP_Free(OscUnit);
    	}
	OscUnit = NULL;
	if (OscBuffer){
		free(OscBuffer);
    	}
	OscBuffer = NULL;
	FSOUND_Close();
	nas_destroy_dialog(nas_amp);
}

static int DialogCallback(NAS_WIDGET *d, int msg, int *w, int *h){
	NAS_WIDGET *tmp = amp_prev;
	int th = alfont_text_height(AL_font);
	if(msg == MSG_END)
		return D_O_K;
	amp_group->w = amp_spect->w + groupbox.rpadding + groupbox.lpadding + 8;
	amp_group->h = amp_spect->h + groupbox.tpadding + groupbox.epadding + th + groupboxtop.tpadding + groupboxtop.epadding + 6;
	amp_volume->x = amp_group->x + amp_group->w + 5;
	amp_volume->y = 12;
	amp_time_box->x = amp_group->x + nasfont_text_length(AL_font, nas_get_label(amp_time_text), 0);
	amp_time_box->y = amp_group->y + amp_group->h + 5;
	amp_time_box->h = th + 4;
	amp_time_box->w = nasfont_text_length(AL_font, "00:00.00", 0) + 4;
	amp_time_text->x = amp_group->x;
	amp_time_text->y = amp_time_box->y + 2;
	amp_cur_time->x = amp_group->x + nasfont_text_length(AL_font, nas_get_label(amp_time_text), 0) + 2;
	amp_cur_time->y = amp_time_box->y + 2;
	amp_cur_time->w = nasfont_text_length(AL_font, "00:00.00", 0);
	amp_kbps_text->x = amp_time_box->x + amp_time_box->w + 5;
	amp_kbps_text->y = amp_time_text->y;
	amp_kbps_box->x = amp_kbps_text->x + alfont_text_length(AL_font, nas_get_label(amp_kbps_text));
	amp_kbps_box->y = amp_time_box->y;
	amp_kbps_box->h = th + 4;
	amp_kbps_box->w = alfont_text_length(AL_font, "000") + 4;
	amp_cur_kbps->x = amp_kbps_box->x + 2;
	amp_cur_kbps->y = amp_kbps_box->y + 2;
	amp_cur_kbps->w = alfont_text_length(AL_font, "000");
	amp_cur_kbps->h = th;
	amp_progress->x = amp_group->x;
	amp_progress->y = amp_time_box->y + amp_time_box->h + 5; // can't be right... check later
	amp_progress->w = amp_group->w;
	amp_shuffle->x = amp_group->x;
	amp_shuffle->y = amp_progress->y + amp_progress->h + 5;
	amp_repeat->x = amp_group->x;
	amp_repeat->y = amp_shuffle->y + amp_shuffle->h + 5;
	amp_prev->x = amp_shuffle->x;
	amp_prev->y = amp_repeat->y + th + 5;
	while(tmp != amp_eject){
		tmp->next->x = tmp->x + tmp->w + 5;
		tmp->y = amp_repeat->y + amp_repeat->h + 5;
		tmp = tmp->next;
	}
	amp_eject->y = amp_repeat->y + amp_repeat->h + 5;
	amp_list->x = amp_volume->x + amp_volume->w + 5;
	if(msg == MSG_START){
		*h = amp_prev->y + amp_prev->h + 2;
		*w = amp_list->x + amp_list->w + 2;
	}
	amp_scroll_box->x = amp_list->x;
	amp_scroll_box->y = amp_group->y;
	amp_scroll_box->w = amp_list->w;
	amp_scroll_box->h = th + 4;
	amp_list->y = amp_scroll_box->y + amp_scroll_box->h + 5;
	amp_list->h = amp_prev->y - amp_list->y - 5;
	amp_scroll_text->x = amp_scroll_box->x + 2;
	amp_scroll_text->y = amp_scroll_box->y + 2;
	amp_scroll_text->w = amp_scroll_box->w - 4;
	amp_scroll_text->h = th;
	return D_O_K;
}

static int get_kbps(FSOUND_STREAM *stream){
	double lenms, lenbytes;
	if(stream){
		lenms = FSOUND_Stream_GetLengthMs(stream);
		lenbytes = FSOUND_Stream_GetLength(stream);
        	return lenbytes / (lenms / 1000) / 1000 * 8;
	}
	return 0;
}

static int GroupboxCallback(NAS_WIDGET *d, void *dp){
	amp_spect->flags ^= AMP_OSS;
	if(amp_spect->flags & AMP_OSS)
		nas_set_label(amp_group, "Oscilloscope");
	else
		nas_set_label(amp_group, "Spectrum");
	return D_O_K;
}

static int nas_spectrum_proc(int msg, NAS_WIDGET *d, BITMAP *wnd, int c){
	SPECTRUM_STATE *state = (SPECTRUM_STATE*)d->dp;
	int minutes = 0, seconds = 0, ms = 0, pos = 0;
	char curtime[9];
	switch(msg){
		case MSG_START:
			if(stream)
				kbps = get_kbps(stream);
			else if(mod)
				kbps = FMUSIC_GetBPM(mod);
			else
				kbps = 0;
		break; 
		case MSG_RESIZE:
			nas_invalidate_canvas(d, d->x, d->y, d->w, d->h);
			if(d->flags & AMP_OSS){
				FSOUND_DSP_SetActive(OscUnit, TRUE);
				PlotOscilliscope(wnd, d->flags, d->x, d->y, d->w, d->h);
			}
			else{
				FSOUND_DSP_SetActive(FSOUND_DSP_GetFFTUnit(), TRUE);
				PlotSpectrum(wnd, state->num_bars, state->falloff_rate, d->flags, d->x, d->y, d->w, d->h);	
			}
		break;
		case MSG_DRAW:
			if(stream)
				ms = FSOUND_Stream_GetTime(stream);
			else if(mod){
				if(FMUSIC_GetType(mod) != FMUSIC_TYPE_MIDI)
					ms = FMUSIC_GetTime(mod);
				else
					ms = 0;
			}
			else if(fsound_cd){
				if(playing)
					ms = FSOUND_CD_GetTrackTime(fsound_cd_drive);
				else
					ms = 0;
			}
			seconds = ms / 1000;
			if(!fsound_cd) 		
				ms = (ms / 10) % 100;
			else
				ms = 0;		//Truncate ms to 0, fsound CD updating is SLOW
			minutes = seconds / 60; 
			seconds %= 60;
			while(game_time > 0){
				nas_invalidate_canvas(d, d->x, d->y, d->w, d->h);
				if(d->flags & AMP_OSS){
					FSOUND_DSP_SetActive(OscUnit, TRUE);
					PlotOscilliscope(wnd, d->flags, d->x, d->y, d->w, d->h);
				}
				else{
					FSOUND_DSP_SetActive(FSOUND_DSP_GetFFTUnit(), TRUE);
					PlotSpectrum(wnd, state->num_bars, state->falloff_rate, d->flags, d->x, d->y, d->w, d->h);	
				}
				game_time--;
			}
			uszprintf(curtime, sizeof(curtime), "%02d:%02d.%02d", minutes, seconds, ms);
			nas_set_label(amp_cur_time, curtime);
		break;
		case MSG_DCLICK:
			if(mouse_b & 1){
				d->flags ^= AMP_OSS;
				if(d->flags & AMP_OSS)
					nas_set_label(amp_group, "Oscilloscope");
				else
					nas_set_label(amp_group, "Spectrum");
			}
		break;
		case MSG_IDLE:
			nas_object_message(d, MSG_DRAW, 0);
			if(stream){
				char percent[5];
				if(fsound_net_stream){
					FSOUND_Stream_Net_GetStatus(stream, 0, &pos, 0, 0);
					pos = (pos / (float)100) * nas_get_progress_maxval(amp_progress);
				}
				else{
					unsigned int lenms = FSOUND_Stream_GetPosition(stream);
					unsigned int len = FSOUND_Stream_GetLength(stream);
					pos = (lenms / (float)len) * nas_get_progress_maxval(amp_progress);
				}
				if(nas_get_progress_position(amp_progress) != pos){
					nas_set_progress_position(amp_progress, pos);
					uszprintf(percent, sizeof(percent), "%d%%", (int)((nas_get_progress_position(amp_progress) / (float)nas_get_progress_maxval(amp_progress)) * 100));
					nas_set_label(amp_progress, percent);
				}
			}
			else if(mod){ 
				if(FMUSIC_GetType(mod) != FMUSIC_TYPE_MIDI){
					char percent[5];
					int num_order = FMUSIC_GetNumOrders(mod);
					int order = FMUSIC_GetOrder(mod);
					pos = nas_get_progress_maxval(amp_progress) * (order / (float)num_order);
					if(nas_get_progress_position(amp_progress) != pos){
						nas_set_progress_position(amp_progress, pos);
						uszprintf(percent, sizeof(percent), "%d%%", (int)((nas_get_progress_position(amp_progress) / (float)nas_get_progress_maxval(amp_progress)) * 100));
						nas_set_label(amp_progress, percent);
					}	
				}
			}
			else if(fsound_cd){
				if(playing){
					char percent[5];
					unsigned int lenms = FSOUND_CD_GetTrackTime(fsound_cd_drive);
					unsigned int len = FSOUND_CD_GetTrackLength(fsound_cd_drive, fsound_cd_track);
					pos = ((lenms / (float)len) * nas_get_progress_maxval(amp_progress));
					if(nas_get_progress_position(amp_progress) != pos){
						nas_set_progress_position(amp_progress, pos);
						uszprintf(percent, sizeof(percent), "%d%%", (int)((nas_get_progress_position(amp_progress) / (float)nas_get_progress_maxval(amp_progress)) * 100));
						nas_set_label(amp_progress, percent);
					}
				} 
			}
		break;
		case MSG_DESTROY:
			if(d->dp)
				free(d->dp);
		break;
	}
	return D_O_K;
}

static int PlayCallback(NAS_WIDGET *d, void *dp) {
    	if(paused){
		paused^=1;
		if(stream)
			FSOUND_SetPaused(0, paused);
		else if(mod)
			FMUSIC_SetPaused(mod, paused);
		else if(fsound_cd)
			FSOUND_CD_SetPaused(fsound_cd_drive, paused);
    	}
    	else if(stream){
    		FSOUND_Stream_Play(0, stream);
   		playing = TRUE;
	}		
	else if(mod && playing == FALSE){
		FMUSIC_PlaySong(mod);
		playing = TRUE;
	}
	else if(fsound_cd){
		FSOUND_CD_SetPlayMode(fsound_cd_drive, FSOUND_CD_PLAYONCE);
		FSOUND_CD_Play(fsound_cd_drive, fsound_cd_track);
		playing = TRUE;
	}
    	return D_O_K;
}

static int StopCallback(NAS_WIDGET *d, void *dp) {
	if(stream)
    		FSOUND_Stream_Stop(stream);	
	else if(mod)
		FMUSIC_StopSong(mod);
	else if(fsound_cd)
		FSOUND_CD_Stop(fsound_cd_drive);
	playing = FALSE;
	paused = 0;
	numplayed = 0;
   	return D_O_K;
}

static int CloseCallback(NAS_WIDGET *d, void *dp){
	if(stream)
		FSOUND_Stream_Stop(stream);
	if(mod)
		FMUSIC_StopSong(mod);
	if(fsound_cd)
		FSOUND_CD_Stop(fsound_cd_drive);
	playing = FALSE;
	paused = 0;
	fsound_net_stream = 0;
	item_playing = NULL;
	return D_CLOSE;
}

static int PauseCallback(NAS_WIDGET *d, void *dp){
	if(playing && !fsound_net_stream){
		paused ^= 1;
		if(stream)
			FSOUND_SetPaused(0, paused);
		else if(mod)
			FMUSIC_SetPaused(mod, paused);
		else if(fsound_cd)
			FSOUND_CD_SetPaused(fsound_cd_drive, paused);
	}
	return D_O_K;
}

static int EjectCallback(NAS_WIDGET *d, void *dp){
	int was = nasmouse_set_mouse_sprite(NASMOUSE_WAIT);
	if((playing  || paused) && fsound_cd)
		FSOUND_CD_Stop(fsound_cd_drive);
	FSOUND_CD_Eject(fsound_cd_drive);
	nasmouse_set_mouse_sprite(was);
	return D_O_K;
}

static int RepeatCallback(NAS_WIDGET *d, void *dp){
	repeat ^= 1;
	if(mod)
		FMUSIC_SetLooping(mod, repeat);
	return D_O_K;
}

static int ShuffleCallback(NAS_WIDGET *d, void *dp){
	shuffle ^=1;
	if(!shuffle)
		numplayed = 0;
	return D_O_K;
}

static int ProgressSetPosCallback(NAS_WIDGET *d, int max, int pos, void *dp){
	if (mod)
		FMUSIC_SetOrder(mod, (int)((float)FMUSIC_GetNumOrders(mod) / (float)max * (float)pos));
	else if (stream && !fsound_net_stream)
        	FSOUND_Stream_SetTime(stream, (int)((float)FSOUND_Stream_GetLengthMs(stream) / (float)max * (float)pos));
	else if(fsound_cd)
		FSOUND_CD_SetTrackTime(fsound_cd_drive, FSOUND_CD_GetTrackLength(fsound_cd_drive, fsound_cd_track) / (float)max * (float)pos);
	return D_O_K;
}

static int PlayerProgressCallback(NAS_WIDGET *d, void *dp){
	AMP_ITEM *amp;
	LIST_ITEM *item;
	int idx, selected;
	int itemcount = nas_get_listbox_numitems(amp_list);
	int flag = 0;
	int numsongs = 0;
	int len;
	char curkbps[4];
	if(fsound_net_stream && item_playing != NULL){
		int flags;
		void *tmp = NULL;
		if(FSOUND_Stream_Net_GetStatus(stream, 0, 0, &kbps, &flags)){
			if (flags & FSOUND_PROTOCOL_SHOUTCAST)
                    		FSOUND_Stream_FindTagField(stream, FSOUND_TAGFIELD_SHOUTCAST, "icy-name", &tmp, 0);
                	else if (flags & FSOUND_PROTOCOL_ICECAST)
                    		FSOUND_Stream_FindTagField(stream, FSOUND_TAGFIELD_ICECAST, "ice-name", &tmp, 0);
		}
		if(tmp && !item_playing->title){
			char buffer[256];
			item_playing->title = (char*)malloc(sizeof(char) * strlen((char*)tmp) + 1);
			ustrcpy(item_playing->title, (char*)tmp);
			nas_object_message(amp_list, MSG_DRAW, 0);
			ustrncpy(buffer, item_playing->title, sizeof(buffer));
			len = alfont_text_length(AL_font, buffer);
			if(len > amp_scroll_text->w)
				strncat(buffer, " *** ", sizeof(buffer));
			nas_set_textscroll_delay(amp_scroll_text, 40);
			nas_set_textscroll_position(amp_scroll_text, 0);
			nas_set_label(amp_scroll_text, buffer);
		}
		if(old_kbps != kbps){
			uszprintf(curkbps, sizeof(curkbps), "%d", kbps);
			nas_set_label(amp_cur_kbps, curkbps);
			old_kbps = kbps;
		}
	}
	if(metanum){
		char buffer[256];
		if(item_playing->title)
			uszprintf(buffer, sizeof(buffer), "%s *** %s - %s", item_playing->title, net_artist, net_title);
		else
			uszprintf(buffer, sizeof(buffer), "%s - %s", net_artist, net_title);
		len = alfont_text_length(AL_font, buffer);
		if(len > amp_scroll_text->w)
			ustrncat(buffer, " *** ", sizeof(buffer));
		if(ustrcmp(buffer, nas_get_label(amp_scroll_text)) != 0){
			nas_set_textscroll_delay(amp_scroll_text, 40);
			nas_set_textscroll_position(amp_scroll_text, 0);
			nas_set_label(amp_scroll_text, buffer);
		}
		metanum = 0;	
        }
	if(shuffle){
		if(mod && FMUSIC_GetType(mod) != FMUSIC_TYPE_MIDI && FMUSIC_IsFinished(mod) && playing == TRUE){ 
			flag = 1;
		}
		else if(stream){
			if(fsound_net_stream){
				int status = FSOUND_Stream_GetOpenState(stream);
				if(status == -1 || status == -3)
					flag = 1;
				else if(status == 0 && playing)
					PlayCallback(d, NULL);
			}
			else if((FSOUND_IsPlaying(0) == 0) && playing == TRUE)
				flag = 1;
		}
		if(flag){
			do{
				flag = 0;
				if(!repeat){
					if(numplayed < itemcount - 1)
						numplayed ++;
					else
						flag = 1;
				}
				if(!flag)
					curstatus = (curstatus + 1) % itemcount;
				for(idx = 0; idx < itemcount; idx++){
					item = nas_get_listbox_item(amp_list, idx);
					amp = item->userobj;
					if(amp->status == curstatus){
						nas_listbox_select(amp_list, idx);
						LoadfileCallback(amp_list, -1, item, NULL);
						nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
						break;
					}
				}
				numsongs++;
			}while(numsongs < itemcount - 1 && !stream && !mod && !fsound_cd);
			if(!flag)
				PlayCallback(d, NULL);
		}
	}
	if(repeat){
		if(mod && FMUSIC_GetType(mod) != FMUSIC_TYPE_MIDI && FMUSIC_IsFinished(mod) && playing == TRUE){
			playing = FALSE;
			PlayCallback(d, NULL);
		}
		else if(stream && (FSOUND_IsPlaying(0) == 0) && playing == TRUE){
			playing = FALSE;
			PlayCallback(d, NULL);
		}
	}
	if(!shuffle && !repeat){
		if(mod && FMUSIC_GetType(mod) != FMUSIC_TYPE_MIDI && FMUSIC_IsFinished(mod) && playing == TRUE)
			flag = 1;
		else if(stream){
			if(fsound_net_stream){
				int status = FSOUND_Stream_GetOpenState(stream);
				if(status == -1 || status == -3)
					flag = 1;
				else if(status == 0 && playing)
					PlayCallback(d, NULL);
			}
			else if((FSOUND_IsPlaying(0) == 0) && playing == TRUE)
				flag = 1;
		}
		if(flag){
			flag = 0;
			do{
				selected = nas_get_listbox_selected(amp_list, NULL);
				if(selected != -1 && selected < itemcount - 1)
					nas_listbox_select(amp_list, selected + 1);
				else{
					playing = FALSE;
					flag = 1;
					break;
				}	
				LoadfileCallback(amp_list, -1, nas_get_listbox_item(amp_list, selected + 1), NULL);
				nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
			}while(selected < itemcount - 1 && !stream && !mod && !fsound_cd);
			if(!flag)
				PlayCallback(d, NULL);
		}
	}
	return D_O_K;
}

static int get_track_num(char *track){
	char buffer[4];
	char *ptr;
	int i = 0;
	ptr = track;
	while(*ptr != '.' && *ptr != '\0')
		ptr++;
	ptr--;
	while(ptr != track && isdigit(*ptr))
		ptr--;
	ptr ++;
	while(*ptr != '.' && *ptr != '\0'){
		buffer[i] = *ptr;
		ptr++;
		i++;
	}
	buffer[i] = '\0';
	return atoi(buffer);
}

static int LoadfileCallback(NAS_WIDGET *d, int msg, LIST_ITEM *item, void *dp){
	if(msg == MSG_DCLICK || msg == -1){
		char curkbps[4];
		AMP_ITEM *amp = (AMP_ITEM *)item->userobj;
		char buffer[256];
		int seconds, minutes, len, was;
		char *ext;
		was = nasmouse_set_mouse_sprite(NASMOUSE_WAIT);
		if(stream || mod || fsound_cd){
			if(playing || paused){
				if(stream)
					FSOUND_Stream_Stop(stream);
				else if(mod)
					FMUSIC_StopSong(mod);
				else if(fsound_cd)
					FSOUND_CD_Stop(fsound_cd_drive);
    				playing = FALSE;
    				paused = 0;
			}
			if(stream){
				FSOUND_Stream_Close(stream);
				stream = NULL;
				playing = FALSE;
			}
			else if(mod){
				FMUSIC_FreeSong(mod);
				mod = NULL;
				playing = FALSE;
			}
			else if(fsound_cd){
				fsound_cd = 0;
				playing = FALSE;
			}
			fsound_net_stream = 0;
			item_playing = NULL;
		}
		if(msg == MSG_DCLICK)
			numplayed = 0;		//Reset the shuffle variable
		curstatus = amp->status;
		ustrcpy(buffer, amp->name);
		ext = get_extension(buffer);
		if(ustrcmp(ext, "cda") == 0){
			fsound_cd = 1;
			fsound_cd_track = get_track_num(get_filename(buffer));
		}
		else
			fsound_cd = 0;
		if(!fsound_cd){
			mod = FMUSIC_LoadSong(buffer);
			if(!mod)
				stream = FSOUND_Stream_Open(buffer, FSOUND_STEREO | FSOUND_2D | FSOUND_16BITS | FSOUND_SIGNED | FSOUND_NONBLOCKING, 0, 0);
		}
		if(stream){
			int test;
			do{
				test = FSOUND_Stream_GetOpenState(stream);
				nasmouse_update_mouse();
			}while(test == -2 && test != 0 && test != -1 && test != -3);
			if(test == -1 || test == -3)
				stream = NULL;
			if(stream){
				if (!ustrncmp(buffer, "http://", 7) || !ustrncmp(buffer, "http:\\\\", 7)){
					fsound_net_stream = 1;
					FSOUND_Stream_Net_SetMetadataCallback(stream, metacallback, 0);
					kbps = 0;
				}
				else{
					fsound_net_stream = 0;
					kbps = get_kbps(stream);
				}
				item_playing = amp;
			}
		}
		else if(mod){
			fsound_net_stream = 0;
			kbps = FMUSIC_GetBPM(mod);
			FMUSIC_SetLooping(mod, repeat);
			item_playing = amp;
		}
		else{
			fsound_net_stream = 0;
			item_playing = NULL;
			kbps = 0;
		}
		uszprintf(curkbps, sizeof(curkbps), "%d", kbps);
		nas_set_label(amp_cur_kbps, curkbps);
		if (mod){
			if (FMUSIC_GetType(mod) != FMUSIC_TYPE_IT)		/* IT has its own master volume setting */
				FMUSIC_SetMasterVolume(mod, 192);      
			if (FMUSIC_GetType(mod) == FMUSIC_TYPE_MOD || FMUSIC_GetType(mod) == FMUSIC_TYPE_S3M)
				FMUSIC_SetPanSeperation(mod, 0.85f);	    /* 15% crossover */
			nas_set_slider_position(amp_volume, nas_get_slider_maxval(amp_volume) - FMUSIC_GetMasterVolume(mod));
		}
		if(stream){
			nas_set_slider_position(amp_volume, nas_get_slider_maxval(amp_volume) - FSOUND_GetVolume(0));
		}
		if(amp->seconds == 0){
			double ms;
			if(stream){
				if (!ustrncmp(buffer, "http://", 7) || !ustrncmp(buffer, "http:\\\\", 7))
					ms = 0;
				else
					ms = FSOUND_Stream_GetLengthMs(stream);
			}
			else if(fsound_cd)
				ms = FSOUND_CD_GetTrackLength(fsound_cd_drive, fsound_cd_track);
			else
				ms = 0;
			amp->seconds = ms / 1000;
		}
		seconds = amp->seconds;
		minutes = seconds / 60; 
		seconds %= 60;
		if(!amp->title)
			make_title(&amp);
		if(fsound_net_stream)
			uszprintf(buffer, sizeof(buffer), "%s", amp->title ? amp->title : amp->name);
		else
			uszprintf(buffer, sizeof(buffer), "%s [%02d:%02d]", amp->title ? amp->title : amp->name, minutes, seconds);
		len = alfont_text_length(AL_font, buffer);
		if(len > amp_scroll_text->w)
			ustrncat(buffer, " *** ", sizeof(buffer));
		nas_set_textscroll_delay(amp_scroll_text, 40);
		nas_set_textscroll_position(amp_scroll_text, 0);
		nas_set_label(amp_scroll_text, buffer);
		nasmouse_set_mouse_sprite(was);
	}
	return D_O_K;
}

static int VolumeCallback(NAS_WIDGET *d, int max, int pos, void *dp){
	if(stream)
		FSOUND_SetVolume(0, max - pos);
	else if(mod)
		FMUSIC_SetMasterVolume(mod, max - pos);
	else if(fsound_cd)
		FSOUND_CD_SetVolume(fsound_cd_drive, max - pos);
	return D_O_K;
}

static int NextCallback(NAS_WIDGET *d, void *dp){
	LIST_ITEM *item;
	AMP_ITEM *amp;
	int was = playing;
	int idx, selected;
	int itemcount = nas_get_listbox_numitems(amp_list);
	int flag = 0;
	int numsongs = 0;
	if(shuffle){
		do{
			flag = 0;
			if(!repeat){
				if(numplayed < itemcount - 1)
					numplayed ++;
				else
					flag = 1;
			}
			else
				numplayed = 0;
			if(!flag)
				curstatus = (curstatus + 1) % itemcount;
			for(idx = 0; idx < itemcount; idx++){
				item = nas_get_listbox_item(amp_list, idx);
				amp = (AMP_ITEM *)item->userobj;
				if(amp->status == curstatus){
					nas_listbox_select(amp_list, idx);
					LoadfileCallback(amp_list, -1, item, NULL);
					nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
					break;
				}
			}
			numsongs++;
		}while(numsongs < itemcount - 1 && !stream && !mod && !fsound_cd);
		if(!flag && was)
			PlayCallback(d, NULL);
	}
	else{
		do{
			if(repeat){
				selected = nas_get_listbox_selected(amp_list, NULL);
				nas_listbox_select(amp_list, (selected + 1) % itemcount);
				numsongs ++;
				if(numsongs == itemcount - 1){
					playing = FALSE;
					flag = 1;
					break;
				}
			}
			else{
				selected = nas_get_listbox_selected(amp_list, NULL);
				if(selected != -1 && selected < itemcount - 1)
					nas_listbox_select(amp_list, selected + 1);
				else{
					playing = FALSE;
					flag = 1;	
					break;
				}
			}
			LoadfileCallback(amp_list, -1, nas_get_listbox_item(amp_list, (selected + 1) % itemcount), NULL);
			nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
		}while(!stream && !mod && !fsound_cd);
		if(!flag && was)
			PlayCallback(d, NULL);
	}
	return D_O_K;
}

static int PrevCallback(NAS_WIDGET *d, void *dp){
	LIST_ITEM *item;
	AMP_ITEM *amp;
	int was = playing;
	int idx, selected;
	int itemcount = nas_get_listbox_numitems(amp_list);
	int flag = 0;
	int numsongs = 0;
	if(shuffle){
		do{
			if(!repeat){
				if(numplayed > 0)
					numplayed --;
				else
					flag = 1;
			}
			else
				numplayed = 0;
			if(!flag){
				curstatus = curstatus - 1;
				if(curstatus < 0)
					curstatus = itemcount - 1;
			}
			for(idx = 0; idx < itemcount; idx++){
				item = nas_get_listbox_item(amp_list, idx);
				amp = (AMP_ITEM *)item->userobj;
				if(amp->status == curstatus){
					nas_listbox_select(amp_list, idx);
					LoadfileCallback(amp_list, -1, item, NULL);
					nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
					break;
				}
			}
			numsongs++;
		}while(numsongs < itemcount - 1 && !stream && !mod && !fsound_cd);
		if(!flag && was)
			PlayCallback(d, NULL);
	}
	else{
		do{
			if(repeat){
				selected = nas_get_listbox_selected(amp_list, NULL);
				if(selected > 0)
					nas_listbox_select(amp_list, selected - 1);
				else
					nas_listbox_select(amp_list, itemcount - 1);
				numsongs ++;
				if(numsongs == itemcount - 1){
					playing = FALSE;
					flag = 1;
					break;
				}
			}
			else{
				selected = nas_get_listbox_selected(amp_list, NULL);
				if(selected > 0)
					nas_listbox_select(amp_list, selected - 1);
				else{
					playing = FALSE;
					flag = 1;	
					break;
				}
			}
			if(selected > 0)
				LoadfileCallback(amp_list, -1, nas_get_listbox_item(amp_list, selected - 1), NULL);
			else
				LoadfileCallback(amp_list, -1, nas_get_listbox_item(amp_list, itemcount - 1), NULL);
			nas_set_listbox_focus(amp_list, NAS_LIST_SEEK_SELECTION);
		}while(!stream && !mod && !fsound_cd);
		if(!flag && was)
			PlayCallback(d, NULL);
	}
	return D_O_K;
}

/*
	if(msg == MSG_IDLE){
		const char *path = get_config_string("AMP", "path", "./");
		if(tracker <= numfiles){
			ustrcpy(buffer, path);
			ustrcat(buffer, media_files[tracker].name);
			//strcpy(buffer, media_files[tracker].name);
			if((fin = fopen(buffer, "rb+")) == NULL){
					return D_O_K;
			}
			getTag(fin, &info);
			if(info.Artist[0] == '\0'){
				ustrcpy(media_files[tracker].artist, getartist(media_files[tracker].name));
				ustrcpy(media_files[tracker].title, gettitle(media_files[tracker].name));
			}
			else{
				ustrcpy(media_files[tracker].artist, info.Artist);
				ustrcpy(media_files[tracker].title, info.Title);
			}
			fclose(fin);
			tracker++;
			nas_object_message(amp_list, MSG_DRAW, 0);
		}
	}
	if(msg == MSG_END){
		destroy_filesys();
		numfiles = 0;
		tracker = 0;
	}
*/

// Courtesy of fmod main.c
static void *F_CALLBACKAPI OscCallback(void *originalbuffer, void *newbuffer, int length, int param){
	int mixertype = FSOUND_GetMixer();
    	int count;
	int totalblocks; 
    	signed short    *dest;
	totalblocks = FSOUND_DSP_GetBufferLengthTotal() / FSOUND_DSP_GetBufferLength(); 
    	/*
        Convert and downmix into a mono short int buffer.
    	*/
    	dest = &OscBuffer[OscBlock * FSOUND_DSP_GetBufferLength()];
    	if (mixertype == FSOUND_MIXER_QUALITY_FPU){
        	float *src = (float *)newbuffer;
        	for (count=0; count < length; count++){
            		dest[count] = (signed short)(src[count << 1] + src[(count << 1) + 1] * 0.5f);
        	}
    	}
    	else{
        	signed short *src = (signed short *)newbuffer;
        	for (count=0; count < length; count++){
            		dest[count] = (signed short)(((int)src[count << 1] + (int)src[(count << 1) + 1]) >> 1);
        	}
    	}    
    	OscBlock++;
    	if (OscBlock >= totalblocks){
        	OscBlock = 0;
    	}
	return newbuffer;
}

// Plots the spectrum bar graph onto buffer
static void PlotSpectrum(BITMAP *dest, int num_bars, int falloff_rate, int flags, int x, int y, int w, int h){
	int i;
	int w2 = w;
	int x1 = x;
	int w1 = (w2 / num_bars);
	int y1 = y + h;
	int bars = num_bars;
	rectfill(dest, x, y, x + w - 1, y + h - 1, makecol(0, 0, 0));
	spectrum = FSOUND_DSP_GetSpectrum();
	for(i = 0; i < num_bars; i++){
		if(bars - i > 0)
			w1 = w2 / (bars - i);
		else
			w1 = 0;
		w2 -= w1;
		if (s[i] < spectrum[i])
			s[i] = spectrum[i];
		else
		   	s[i] = s[i] - (s[i] - spectrum[i]) / falloff_rate;
		if(flags & AMP_LINE) 
			draw_bar_gradient(dest, x1, y + h - 1, x1 + w1 - 1, spectrum[i] * h, spectrum[i] * h, amp_col2, amp_col1);
		else if(flags & AMP_FIRE)
			draw_bar_gradient(dest, x1, y + h - 1, x1 + w1 - 1, spectrum[i] * h, spectrum[i] * h, amp_col1, amp_col2);
		else
			draw_bar_gradient(dest, x1, y + h - 1, x1 + w1 - 1, spectrum[i] * h, h, amp_col1, amp_col2);
		if(flags & AMP_PEAKS)
			if(s[i] * h > 1)
				hline(dest, x1, y1 - (int)(s[i] * h), x1 + w1 - 1, amp_col3);
		x1 += w1;
	}
}

// Draws a gradient bar for the sepectrum analyzer from col1 to col2, h and h2 determine a firescope from a normal scope
static void draw_bar_gradient(BITMAP *dest, int x, int y, int w, int h, int h2, int col1, int col2){
	int r1, r2, g1, g2, b1, b2, i, col3;
	float r_grad, b_grad, g_grad;
	r1 = getr(col1);
	r2 = getr(col2);
	g1 = getg(col1);
	g2 = getg(col2);
	b1 = getb(col1);
	b2 = getb(col2);
	r_grad = ((r1 - r2) / (float)h2);
	b_grad = ((b1 - b2) / (float)h2);
	g_grad = ((g1 - g2) / (float)h2);
	for(i = 0; i < h; i++){
		col3 = makecol(r1 - (i * r_grad), g1 - (i * g_grad), b1 - (i * b_grad));
		hline(dest, x, y - i, w, col3);
	}
}

// Courtesy Copy & Paste from fmod main.c with minor changes
static void PlotOscilliscope(BITMAP *dest, int flags, int x, int y, int w, int h){ 
    	if (OscBuffer){
        	int count, count2, offset;
		int r1, r2, g1, g2, b1, b2, col3;
		float r_grad, b_grad, g_grad;
        	float xoff, step;
		signed short *src;
		int h2 = h >> 1;
		r1 = getr(amp_col1);
		r2 = getr(amp_col2);
		g1 = getg(amp_col1);
		g2 = getg(amp_col2);
		b1 = getb(amp_col1);
		b2 = getb(amp_col2);
		r_grad = ((r1 - r2) / (float)h2);
		b_grad = ((b1 - b2) / (float)h2);
		g_grad = ((g1 - g2) / (float)h2);
		rectfill(dest, x, y, x + w - 1, y + h - 1, makecol(0, 0, 0));	
        	/*
            	The next pcmblock (Oscblock + 1) is the one that is audible.
        	*/
        	offset = (OscBlock + 1) * FSOUND_DSP_GetBufferLength();
        	if (offset >= FSOUND_DSP_GetBufferLengthTotal()){
            		offset -= FSOUND_DSP_GetBufferLengthTotal();
        	}
       		src = &OscBuffer[offset];
        	/*
            	xoff is the x position that is scaled lookup of the dsp block according to the graphical
            	window size.
        	*/
        	xoff = 0;
        	step = (float)FSOUND_DSP_GetBufferLength() / (float)w;
        	for (count=0; count < w; count++){
            		int xx, yy, y2;
            		xx  = (int)xoff;
            		yy  = (int)(((float)src[xx] + 32768.0f) / 65536.0f * (float)h);
            		y2 = (int)(((float)src[xx+(int)step] + 32768.0f) / 65536.0f * (float)h);
            		if (yy > y2){
                		int tmp = yy;
                		yy = y2;
                		y2 = tmp;
            		}
			if(flags & AMP_FIRE){
            			for (count2 = yy; count2 <= y2; count2++){
					col3 = makecol(r1 - (abs((count2 - h2)) * r_grad), g1 - (abs((count2 - h2)) * g_grad), b1 - (abs((count2 - h2)) * b_grad));
					putpixel(dest, x + count, y + count2, col3);
            			}
			}
			else{
				for (count2=yy; count2 <= y2; count2++)
					putpixel(dest, x + count, y + count2, amp_col1);
			}
            		xoff += step;
        	}
    	}
}
